#include <vector>
#include <map>
#include <iostream>
#include <list>
#include <utility>
#include <string>
#include <iomanip>
#include <set>
#include <fstream>
#include <stdlib.h>

#include "graph.h"
#include "louvainlevel.h"

using namespace std;

double modularity(Graph & g, map<int, int> n2c)
{
    double mod = 0.;
    for(int time = g.mintime(); time < g.maxtime(); time++)
    {
        int L = 0;
        map<int, int> inlinks;
        map<int, int> mdegrees;
        for(std::list< int >::const_iterator it = g.nodes().begin(); it != g.nodes().end(); it++)
        {
            int src = *it;
            if(g.has_node(src, time))
            {
                map<int, int > neighs = g.get_neighbors(src, time);

                for(std::map<int, int >::iterator itn = neighs.begin(); itn != neighs.end(); itn++ )
                {
                    int dst = itn->first;
                    int weigh = itn->second;
                    if(n2c[src] == n2c[dst] && src <= dst)
                    {
                        if(inlinks.find(n2c[src])==inlinks.end())
                            inlinks[n2c[src]] = weigh;
                        else
                            inlinks[n2c[src]] += weigh;
                    }

                    if(src <= dst)
                    {
                        L += weigh;
                    }

                    if(src == dst) weigh *= 2;
                    if(mdegrees.find(n2c[src])==mdegrees.end())
                        mdegrees[n2c[src]] = weigh;
                    else
                        mdegrees[n2c[src]] += weigh;

                }
            }
        }
        for(map<int, int>::iterator itdeg = mdegrees.begin(); itdeg != mdegrees.end(); itdeg++)
        {
            int com = itdeg->first;
            int deg = itdeg->second;
            map<int, int>::iterator itin = inlinks.find(com);
            int in = 0;
            if(itin != inlinks.end())
            {
                in = itin->second;
            }
            mod += g.duration(time)*((double)in/(double)L - (double(deg)/(2.*(double)L))*(double(deg)/(2.*(double)L)));
        }
    }

    return mod;
}





map<int, int> louvain(Graph & g, bool debug = true)
{
    map<int, int> n2c;

    for(list< int >::const_iterator it = g.nodes().begin(); it != g.nodes().end(); it++)
        n2c[*it] = *it;

    Graph * cur = &g;
//     cur = new Graph();
//     cur->create_from(n2c, g);
    bool cont = true;
    long double oldmod = modularity(g, n2c);
    while(cont)
    {

        cont = false;

        LouvainLevel com;
        com.init(cur);
        com.debug = debug;
        com.one_level();

        for(std::list< int >::const_iterator it = g.nodes().begin(); it != g.nodes().end(); it++)
        {
            n2c[*it] = com.n2c(n2c[*it]);
        }


        long double mod = modularity(*cur, com.m_n2c);
        if(oldmod < mod)
        {
            cont = true;
            oldmod = mod;
            Graph * prev = cur;
            cur = new Graph();
            cur->create_from(com.m_n2c, *prev);
            if(prev != &g)
                delete prev;
        }
        if(debug)
            cerr << std::fixed << std::setw(5) << std::setprecision(5) << "modularity " << mod <<  endl;
    }
    if(cur != &g)
        delete cur;
    return n2c;
}


void display_part(map<int, int> part)
{
    for(std::map< int, int >::const_iterator it = part.begin(); it != part.end(); it++)
    {
        cout << it->first << " " << it->second << endl;
    }
}







