

#include <set>
#include <iostream>
#include <fstream>

#include "graph.h"

using namespace std;


const std::list< int > & Graph::nodes() const
{
    return m_allnodes;
}


void Graph::set_maxtime(int newtime)
{
    m_maxtime = newtime;
}

void Graph::set_mintime(int newtime)
{
    m_mintime = newtime;
}

void Graph::print(int time)
{

//     net neighs = m_neighbors[time];
//     cout << "Totw : " << m_total_weight[time] << " maxtime : " << m_maxtime << " mintime " << m_mintime << " nodes : " << m_allnodes.size() << endl;
//     for(net::const_iterator ita = neighs.begin(); ita != neighs.end(); ita++)
//     {
//         cout << "deg " << ita->first << " " << m_gdegree[time][ita->first] << endl;
//     }
//     for(net::const_iterator ita = neighs.begin(); ita != neighs.end(); ita++)
//     {
//         for(map< int, int >::const_iterator itn = ita->second.begin(); itn != ita->second.end(); itn++ )
//         {
//             cout << ita->first << " " << itn->first << " " << itn->second << endl;
//         }
//     }
//     for(int i = m_mintime; i < m_maxtime; i++)
//         cout << "duration " << i << " " << m_duration[i] << endl;
}

//turn into into the graph between communities of origin
void Graph::create_from(map<int, int> &n2c, const Graph & origin)
{
    m_total_weight.clear();
    m_gdegree.clear();
    m_neighbors.clear();
    m_duration.clear();
    m_allnodes.clear();

    m_total_weight.resize(origin.maxtime());
    m_gdegree.resize(origin.maxtime());
    m_neighbors.resize(origin.maxtime());
    m_duration.resize(origin.maxtime());
    m_maxtime = origin.maxtime();
    m_mintime = origin.mintime();
    set< int > mylist;
    for(list< int >::const_iterator it = origin.nodes().begin(); it != origin.nodes().end(); it++)
    {
        mylist.insert( n2c[*it] );
    }
    int mnode = 0;

    for(set< int >::iterator it = mylist.begin(); it != mylist.end(); it++)
    {
        int node = *it;
        if(node > mnode)
            mnode = node;
        m_allnodes.push_back(node);
        for(int time = 0; time < origin.maxtime(); time++)
        {
            m_gdegree[time][node] = 0;
        }
    }
    for(int time = 0; time < origin.maxtime(); time++)
    {
        m_total_weight[time] = 0;
        m_duration[time] = origin.duration(time);
        m_neighbors[time] = vector< map< int, int > >(mnode+1);
        for(list< int >::const_iterator it = origin.nodes().begin(); it != origin.nodes().end(); it++)
        {
            int src = *it;
            if(origin.has_node(src, time))
            {
                const map< int, int > & neighs = origin.get_neighbors(src, time);
                map< int, int > weighs;
                for(map< int, int >::const_iterator itn = neighs.begin(); itn != neighs.end(); itn++ )
                {
                    int dst = itn->first;
                    int weigh = itn->second;
                    int comsrc = n2c[src];
                    int comdst = n2c[dst];
                    if(dst <= src)
                    {
                        m_total_weight[time] += weigh;
                        m_gdegree[time][comsrc] += weigh;
                        m_gdegree[time][comdst] += weigh;

                        if(m_neighbors[time][comsrc].find(comdst) == m_neighbors[time][comsrc].end())
                            m_neighbors[time][comsrc][comdst] = weigh;
                        else
                            m_neighbors[time][comsrc][comdst] += weigh;

                        if(comsrc != comdst)
                        {
                            if(m_neighbors[time][comdst].find(comsrc) == m_neighbors[time][comdst].end())
                                m_neighbors[time][comdst][comsrc] = weigh;
                            else
                                m_neighbors[time][comdst][comsrc] += weigh;
                        }
                    }
                }
            }
        }
    }
//     print(0);
}

const bool Graph::has_node(int node, int time) const
{
    try
    {
        return !m_neighbors[time].at(node).empty();
    }
    catch(...)
    {
        return false;
    }

    return false;

}

int Graph::degree( const int node, const  int time)
{
    const int test = int(m_gdegree[time][node]);
    return test;
}

const int Graph::maxtime() const
{
    return m_maxtime;
}

const int Graph::mintime() const
{
    return m_mintime;
}

const map<int, int > &  Graph::get_neighbors(int node, int time) const
{
    return m_neighbors[time].at(node);
}


const int Graph::duration(int time) const
{
    return m_duration[time];
}



void Graph::load(std::string fname)
{
    //read maxtime
    ifstream f(fname.c_str());
    f >> m_maxtime;
    m_mintime = 0;
    m_neighbors.resize(m_maxtime);
    m_duration.resize(m_maxtime);
    m_total_weight.resize(m_maxtime);
    m_gdegree.resize(m_maxtime);

    //read nodelist
    int count = 0;
    int mnode = 0;

    while(f.good())
    {
        int val1;
        f >> val1;
        if(val1 < 0)
        {
            int duration_g = 1;
            if( val1 < -1)
            {
                f >> duration_g;
                m_neighbors[0].resize(mnode+1);
                m_duration[0] = (double) duration_g;
            }
            else
            {
                m_neighbors[0].resize(mnode+1);
                m_duration[0] = 1.;
            }
            break;
        }
        if(val1 > mnode)
            mnode = val1;
        m_allnodes.push_back(val1);
        count++;
    }
    int time = 0;
    //read graphs
    while(f.good())
    {
        int val1;
        f >> val1;
        if(val1 < 0)
        {
            time++;
            int duration_g = 1;
            if( val1 < -1)
                f >> duration_g;
            m_total_weight[time] = 0;
            m_duration[time] = (double)duration_g;
            m_neighbors[time].resize(mnode+1);
        }
        else
        {
            int src = val1;
            int dst;
            f >> dst;
            m_gdegree[time][src]++;
            m_gdegree[time][dst]++;
            m_total_weight[time]++;
            m_neighbors[time][src].insert(make_pair(dst, 1));
            m_neighbors[time][dst].insert(make_pair(src, 1));
        }
    }
    if(time != m_maxtime -1)
    {
        cerr << "Something wrong, not enough graphs loaded" << time << " " << m_maxtime << endl;
    }
}


void Graph::read_cin()
{
    //read maxtime
    cin >> m_maxtime;
    m_mintime = 0;
    m_neighbors.resize(m_maxtime);
    m_duration.resize(m_maxtime);
    //read nodelist
    int count = 0;
    while(cin.good())
    {
        int val1;
        cin >> val1;
        if(val1 < 0)
        {
            int duration_g = 1;
            if( val1 < -1)
            {
                cin >> duration_g;
                m_duration[0] = (double) duration_g;
            }
            else m_duration[0] = 1.;
            break;
        }

        m_allnodes.push_back(val1);
        count++;
    }
    int time = 0;
    //read graphs
    while(cin.good())
    {
        int val1;
        cin >> val1;
        if(val1 < 0)
        {
            time++;
            int duration_g = 1;
            if( val1 < -1)
                cin >> duration_g;
            m_duration[time] = (double)duration_g;
        }
        else
        {
            int src = val1;
            int dst;
            cin >> dst;
            m_neighbors[time][src].insert(make_pair(dst, 1));
            m_neighbors[time][dst].insert(make_pair(src, 1));
        }
    }
    if(time != m_maxtime -1)
    {
        cerr << "Something wrong, not enough graphs loaded" << time << " " << m_maxtime << endl;
    }
}