

#include <vector>
#include <map>
#include <iostream>
#include <list>
#include <utility>
#include <string>
#include <iomanip>
#include <set>
#include <fstream>
#include <stdlib.h>
#include <string.h>
#include <queue>
#include "graph.h"
#include "louvainlevel.h"
#include "com_dyn.h"


using namespace std;


Graph glob_g;
int remaining = 0;
int skip = 1;
int ign = 0;
int uniq = 0;
int limit_dur = -1;
bool connex = false;
int limit_update = 500;

int max_node;

class tree_node
{
public :
    int id;
    list<int> times;
    vector<int> n2c;
    int begin;
    int end;
    int duration;
};

class PairNode
{
  public:
    PairNode(int x, int y, double m) { id_t1 = x; id_t2 = y; modularity = m; };
    bool operator<(const PairNode&) const;              //overloaded < operator
    int id_t1;
    int id_t2;
    double modularity;

};


vector<tree_node> nodes;


/* overload the less-than operator so priority queues know how to compare two Height objects */
bool PairNode::operator<(const PairNode& right) const
{
  return modularity < right.modularity;
}



double duration(int mintime, int maxtime)
{
    double dur = 0.;
    for(int time = mintime; time < maxtime; time++)
    {
        dur += glob_g.duration(time);
    }

    return dur;
}


double duration_time(tree_node & val)
{
    double dur = 0.;
    for(list<int>::iterator it = val.times.begin(); it != val.times.end(); it++)
    {
        dur += glob_g.duration(*it);
    }

    return dur;
}

double modularity_time2(tree_node & src_node, tree_node & dst_node)
{
    double mod = 0.;
    for(list<int>::iterator it_time = dst_node.times.begin(); it_time != dst_node.times.end(); it_time++)
    {
        int time = *it_time;
        int L = 0;
        map<int, int> inlinks;
        map<int, int> mdegrees;
        for(std::list< int >::const_iterator it = glob_g.nodes().begin(); it != glob_g.nodes().end(); it++)
        {
            int src = *it;
            if(glob_g.has_node(src, time))
            {
                map<int, int > neighs = glob_g.get_neighbors(src, time);

                for(std::map<int, int >::iterator itn = neighs.begin(); itn != neighs.end(); itn++ )
                {
                    int dst = itn->first;
                    int weigh = itn->second;
                    if(src_node.n2c[src] == src_node.n2c[dst] && src <= dst)
                    {
                        if(inlinks.find(src_node.n2c[src])==inlinks.end())
                            inlinks[src_node.n2c[src]] = weigh;
                        else
                            inlinks[src_node.n2c[src]] += weigh;
                    }

                    if(src <= dst)
                    {
                        L += weigh;
                    }

                    if(src == dst) weigh *= 2;
                    if(mdegrees.find(src_node.n2c[src])==mdegrees.end())
                        mdegrees[src_node.n2c[src]] = weigh;
                    else
                        mdegrees[src_node.n2c[src]] += weigh;

                }
            }
        }
        for(map<int, int>::iterator itdeg = mdegrees.begin(); itdeg != mdegrees.end(); itdeg++)
        {
            int com = itdeg->first;
            int deg = itdeg->second;
            map<int, int>::iterator itin = inlinks.find(com);
            int in = 0;
            if(itin != inlinks.end())
            {
                in = itin->second;
            }
            mod += glob_g.duration(time)*((double)in/(double)L - (double(deg)/(2.*(double)L))*(double(deg)/(2.*(double)L)));
        }
    }
    return mod;
}


double modularity_time(tree_node & src_node, tree_node & dst_node)
{
    double mod = 0.;
    for(list<int>::iterator it_time = dst_node.times.begin(); it_time != dst_node.times.end(); it_time++)
    {
        int time = *it_time;
        int L = 0;
        map<int, int> inlinks;
        map<int, int> mdegrees;
        for(unsigned int src = 0; src < glob_g.m_neighbors[time].size(); ++src)
        {
            if(glob_g.has_node(src, time))
            {
                int n2csrc = src_node.n2c[src];
                map<int, int > neighs = glob_g.get_neighbors(src, time);

                int degree = glob_g.degree(src, time);
                if(mdegrees.find(n2csrc)==mdegrees.end())
                    mdegrees[n2csrc] = degree;
                else
                    mdegrees[n2csrc] += degree;

                for(std::map<int, int >::iterator itn = neighs.begin(); itn != neighs.end(); itn++ )
                {
                    unsigned int dst = itn->first;
                    int weigh = itn->second;
                    if(src < dst)
                    {
                        if(n2csrc == src_node.n2c[dst])
                        {
                            if(inlinks.find(n2csrc)==inlinks.end())
                                inlinks[n2csrc] = weigh;
                            else
                                inlinks[n2csrc] += weigh;
                        }
                        L += weigh;
                    }
                }
            }
        }
        for(map<int, int>::iterator itdeg = mdegrees.begin(); itdeg != mdegrees.end(); itdeg++)
        {
            int com = itdeg->first;
            int deg = itdeg->second;
            map<int, int>::iterator itin = inlinks.find(com);
            int in = 0;
            if(itin != inlinks.end())
            {
                in = itin->second;
            }
            mod += glob_g.duration(time)*((double)in/(double)L - (double(deg)/(2.*(double)L))*(double(deg)/(2.*(double)L)));
        }
    }
    return mod;
}


//turn into into the graph between communities of origin
Graph create_from(list<int>& times)
{

    Graph g;
    g.m_total_weight.clear();
    g.m_gdegree.clear();
    g.m_neighbors.clear();
    g.m_duration.clear();
    g.m_allnodes.clear();

    g.m_total_weight.resize(times.size());
    g.m_gdegree.resize(times.size());
    g.m_neighbors.resize(times.size());
    g.m_duration.resize(times.size());

    g.m_maxtime = times.size();
    g.m_mintime = 0;
    int mnode = 0;
    for(list< int >::const_iterator it = glob_g.nodes().begin(); it != glob_g.nodes().end(); it++)
    {
        int node = *it;
        if(node > mnode)
            mnode = node;
        g.m_allnodes.push_back(node);
        for(int time = 0; time < g.m_maxtime; time++)
        {
            g.m_gdegree[time][node] = 0;
        }
    }
    int new_time = -1;
    for(std::list< int >::const_iterator it = times.begin(); it != times.end(); it++)
    {
        new_time++;
        int time = *it;
        g.m_total_weight[new_time] = 0;
        g.m_duration[new_time] = glob_g.duration(time);
        g.m_neighbors[new_time] = vector< map< int, int > >(mnode+1);
        for(list< int >::const_iterator it = glob_g.nodes().begin(); it != glob_g.nodes().end(); it++)
        {
            int src = *it;
            if(glob_g.has_node(src, time))
            {
                const map< int, int > & neighs = glob_g.get_neighbors(src, time);
                map< int, int > weighs;
                for(map< int, int >::const_iterator itn = neighs.begin(); itn != neighs.end(); itn++ )
                {
                    int dst = itn->first;
                    int weigh = itn->second;
                    g.m_total_weight[new_time] += weigh;
                    g.m_gdegree[new_time][src] += weigh;
                    if(g.m_neighbors[new_time][src].find(dst) == g.m_neighbors[new_time][src].end())
                    {
                        g.m_neighbors[new_time][src][dst] = weigh;
                    }
                    else
                    {
                        g.m_neighbors[new_time][src][dst] += weigh;
                    }
                }
            }
        }
    }
    return g;
}


vector<int> temp_louvain(tree_node t)
{
    Graph g = create_from(t.times);
    map<int, int> n2c = louvain(g, false);
    vector<int> ret(max_node+1, -1);
    for(map<int, int>::iterator it = n2c.begin(); it != n2c.end(); it++)
        ret.at(it->first) = it->second;
    return ret;
}

double similarity(tree_node & times1, tree_node & times2)
{

    double mod1 = modularity_time(times1, times2)/times2.duration;
    double mod2 = modularity_time(times2, times1)/times1.duration;
    return mod1+mod2;;    
}

void exp()
{
    cerr << "Computing initial snapshot communities" << endl;
    //first init, put all the graphs in the possible
    priority_queue<PairNode>  * pq = new priority_queue<PairNode>;
    uniq = 0;
    int real_time = 0;
    for(int time = glob_g.mintime(); time < glob_g.maxtime(); time++)
    {
        cerr << time << endl;
        int myid = uniq;
        uniq++;
        nodes[myid].times.push_back(time);
        nodes[myid].id = myid;
        nodes[myid].n2c = temp_louvain(nodes[myid]);
        nodes[myid].begin = real_time;
        real_time += glob_g.duration(time);
        nodes[myid].end = real_time;
        nodes[myid].duration = glob_g.duration(time);

    }

    cerr << "done, computing inter-snapshot similarities" << endl;

    int numberofnodes = glob_g.maxtime() - glob_g.mintime();
    double remaining = numberofnodes/100.;
    double mcount = 0;
    for(vector<tree_node>::iterator it1 = nodes.begin(); it1 != nodes.end(); it1++)
    {
        if(it1->id >= 0)
        {
            if(remaining < 0)
            {
                    std::cerr << "\b\b\b\b" << std::fixed << std::setw(3) << std::setprecision(1) << 100.0*(mcount/numberofnodes) << endl;
                remaining = numberofnodes/100.;
            }
            remaining--;
            mcount++;
            for(vector<tree_node>::iterator it2 = nodes.begin(); it2 != nodes.end(); it2++)
            {
                if(it1->id < it2->id)
                {
                    if(limit_dur == -1 || (it2->begin >= it1->begin - limit_dur && it2->begin <= it1->end + limit_dur) || (it2->end >= it1->begin - limit_dur && it2->end <= it1->end + limit_dur)  )
                    {
                        if(!connex || (it1->end == it2->begin || it1->begin == it2->end) )
                        {
                            double sim = similarity(*it1, *it2);

                            if(sim > 0.)
                                pq->push(PairNode(it1->id, it2->id, sim));
                        }
                    }
                }
            }
        }
    }
    cerr << "Initialization done, computing the trees" << endl;

    int count = glob_g.maxtime() - glob_g.mintime();
    int done = 0;
    while(pq->size() > 1)
    {
        cerr << "to do : " << count << endl;
        count--;
        done++;

        //First, find the two best matches

        //we pop until an acceptable match, ie a pair whose both elements are not marqued
        bool found = false;
        while(!found && !pq->empty())
        {
            PairNode p = pq->top();
            if(nodes[p.id_t1].id > 0 && nodes[p.id_t2].id > 0)
            {
                found = true;
            }
            else pq->pop();
        }

        if(!found)
        {
            //we do not find any good match, maybe due to limit ?
            cerr << "No match, end" << endl;
            delete pq;
            return;
        }
        else
        {
            PairNode p = pq->top();
            pq->pop();

            for(list<int>::iterator it = nodes[p.id_t1].times.begin(); it != nodes[p.id_t1].times.end(); it++)
            {
                cout << *it << ",";
            }
            cout << " ";

            for(list<int>::iterator it = nodes[p.id_t2].times.begin(); it != nodes[p.id_t2].times.end(); it++)
            {
                cout << *it << ",";
            }
            cout << endl;


            //Then merge them
            int myid = uniq;
            uniq++;

            if(nodes[p.id_t1].begin > nodes[p.id_t2].begin)
                nodes[myid].begin = nodes[p.id_t2].begin;
            else
                nodes[myid].begin = nodes[p.id_t1].begin;


            if(nodes[p.id_t1].end < nodes[p.id_t2].end)
                nodes[myid].end = nodes[p.id_t2].end;
            else
                nodes[myid].end = nodes[p.id_t1].end;

            nodes[myid].times.merge(nodes[p.id_t2].times);
            nodes[myid].times.merge(nodes[p.id_t1].times);

            nodes[myid].duration = nodes[p.id_t2].duration+nodes[p.id_t1].duration;
            nodes[myid].n2c = temp_louvain(nodes[myid]);
            nodes[myid].id = myid;


            nodes[p.id_t2].id = -1;
            nodes[p.id_t2].n2c.clear();
            nodes[p.id_t2].times.clear();

            nodes[p.id_t1].id = -1;
            nodes[p.id_t1].n2c.clear();
            nodes[p.id_t1].times.clear();

            cerr << "Updating similarities" << endl;


            int todo = count;
            int remaining = todo/100.;
            mcount = 0;
            for(vector<tree_node>::iterator it = nodes.begin(); it != nodes.end(); it++)
            {

                if(it->id > 0 && it->id != myid && (limit_dur == -1 || (nodes[myid].begin >= it->begin - limit_dur && nodes[myid].begin <= it->end + limit_dur) || (nodes[myid].end >= it->begin - limit_dur && nodes[myid].end <= it->end + limit_dur)  ))
                {
                    if(!connex || (it->end == nodes[myid].begin || it->begin == nodes[myid].end) )
                    {
                        double sim = similarity(*it, nodes[myid]);
                        if(sim > 0.)
                            pq->push(PairNode(it->id, myid, sim));
                    }
                    if(remaining < 0)
                    {
                        std::cerr << "\b\b\b\b" << std::fixed << std::setw(3) << std::setprecision(1) << 100.0*(mcount/todo) << endl;
                        remaining = todo/100.;
                    }
                    remaining--;
                    mcount++;
                }
            }
            cerr << "Done"  << endl;

            //a bit of cleaning, after some times, when there are many snapshots, the queue containes many time windows that no longer exist, so we clean it.
            if(done == limit_update)
            {
                cerr << "Cleaning priority_queue..." << endl;
                priority_queue<PairNode> * npq = new priority_queue<PairNode>();
                while(!pq->empty())
                {
                    PairNode p = pq->top();
                    if(nodes[p.id_t1].id > 0 && nodes[p.id_t2].id > 0)
                    {
                        npq->push(p);
                    }
                    pq->pop();
                }
                delete pq;
                pq = npq;
                done = 0;
            }
        }

    }

    delete pq;
}


int main(int argc, char** argv)
{
    glob_g.load(argv[1]);
    cerr << "Graph loaded" << endl;

    //This is a hack to increase memory efficiency
    max_node = 0;
    for(std::list< int >::const_iterator it = glob_g.nodes().begin();it != glob_g.nodes().end(); it++)
        if(*it > max_node)
            max_node = *it;
    
    for(int curarg = 2; curarg < argc; )
    {
        if(strcmp(argv[curarg], "--connex")==0)
        {
            connex = true;
            curarg += 1;
            continue;
        }
    }
    remaining = glob_g.maxtime() - glob_g.mintime();

    nodes.resize(2*remaining);
    for(vector<tree_node>::iterator it = nodes.begin(); it != nodes.end(); it++)
        it->id = -1;
    cerr << "Starting " << endl;
    exp();
}


