
#include "louvainlevel.h"
#include <iostream>
#include <vector>
#include <map>
#include <string>
#include <iomanip>

using namespace std;



inline void LouvainLevel::remove(int node, int comm, int dnodecomm, int time)
{
  degree[time][comm] -= m_g->degree(node, time);
  m_n2c[node]  = -1;
}

inline void LouvainLevel::insert(int node, int comm, int dnodecomm, int time)
{
  degree[time][comm] += m_g->degree(node, time);
  m_n2c[node]=comm;
}

void LouvainLevel::one_level()
{
    int numberofnodes = m_g->nodes().size();
    int num = 0;
    int countmove =0;
    bool modif = true;
    int remaining = numberofnodes/100;

    while(modif)
    {
        if(debug)cerr << "Pass " << num++ << " " << countmove <<endl;
        modif = 0;
        countmove = 0;
        double mcount = 0.;
        for(std::list< int >::const_iterator itnodes = m_g->nodes().begin(); itnodes != m_g->nodes().end(); itnodes++)
        {

            mcount += 1.;
            if(remaining == 0)
            {
                 if(debug)std::cerr << "\b\b\b\b" << std::fixed << std::setw(3) << std::setprecision(1) << 100.0*(mcount/numberofnodes) << endl;
                remaining = numberofnodes/100.;
            }
            remaining--;

            int node = *itnodes;
            int com_node = m_n2c[node];
//             cerr << "Considering " << node << endl;
            map< int, vector<int> >  neigh_communities = neighcoms(node);


            map< int, vector<int> >::iterator it = neigh_communities.find(com_node);
            if(it == neigh_communities.end())
            {
                neigh_communities[com_node] = vector<int>(m_g->maxtime(), 0);
            }

            removes(node, com_node, neigh_communities[com_node]);

            int best_com = best_increase(node, com_node, neigh_communities);

            inserts(node, best_com, neigh_communities[best_com]);
//             cerr << "Moving " << node << " from " << com_node << " to " << best_com << endl;
            if(best_com != com_node)
            {
                modif = true;
                countmove++;
            }
        }
    }
}


inline map<int, int> LouvainLevel::neighcom(int node, int time)
{
    map<int, int> result;
    std::map<int, int > neighs = m_g->get_neighbors(node, time);
    for(std::map<int, int >::iterator it1 = neighs.begin(); it1 != neighs.end(); it1++)
    {
        int neighbor = it1->first;
        int weight = it1->second;
        if(neighbor != node)
        {
            if(result.find(m_n2c[neighbor]) == result.end())
                result[m_n2c[neighbor]] = 0;
            result[m_n2c[neighbor]] = result[m_n2c[neighbor]] + weight;
        }
    }
    return result;
}



map< int, vector<int> > LouvainLevel::neighcoms(int node)
{
    vector< map<int, int> > weights(m_g->maxtime());
    for(int time = m_g->mintime(); time < m_g->maxtime(); time++)
    {
        if(m_g->has_node(node, time))
        {
            weights[time] = neighcom(node, time);
        }
    }

    map< int, vector<int> > result;
    for(int time = m_g->mintime(); time < m_g->maxtime(); time++)
    {
        map<int, int> dncs = weights[time];
        for(map<int, int>::iterator it = dncs.begin(); it != dncs.end(); it++)
        {
            int com = it->first;
            int deg = it->second;
            map< int, vector<int> >::iterator itsearch = result.find(com);
            if(itsearch == result.end())
            {
                result[com] = vector<int>(m_g->maxtime(), 0);
            }
            result[com][time] = result[com][time] + deg;
        }

    }
    return result;
}


void LouvainLevel::removes(int& node, int& com, std::vector<int> & weights)
{
//      cout <<"removes" << node << " " << com << " " << weights.size()  << " " << m_g->mintime() << " " << m_g->maxtime() << endl;
    for(int time = m_g->mintime(); time < m_g->maxtime(); time++)
    {
        remove(node, com, weights[time], time);
    }
    m_n2c[node] = -1;
}


void LouvainLevel::inserts(int node, int com, std::vector<int> & weights)
{
//     cout <<"removes" << node << " " << com << " " << weights.size()  << endl;
    m_n2c[node] = com;
    for(int time = m_g->mintime(); time < m_g->maxtime(); time++)
    {
        insert(node, com, weights[time],time);
    }
}


inline int LouvainLevel::best_increase(int node, int com_node, map< int, vector<int> > neigh_communities)
{
    int best_com = com_node;
    double best_increase = 0.;
    for(map< int, vector<int> >::iterator it1 = neigh_communities.begin(); it1 != neigh_communities.end(); it1++)
    {
        int com = it1->first;
        vector<int> dncs = it1->second;
        double incr = 0.;
        for (int time = m_g->mintime(); time < m_g->maxtime(); time++)
        {

            double dnc = (double)dncs[time];
            double totw = (double)m_g->m_total_weight[time];
            double totc = (double)degree[time][com];
            double degc = (double)m_g->degree(node, time);
//              cerr << node << " " << com << " " << dnc << " " << totw << " " << totc << " " << degc << endl;
            if(totw > 0)
                incr +=  m_g->duration(time)*(dnc - totc * degc / (2.*totw));
        }
        if(incr > best_increase)
        {
            best_increase = incr;
            best_com = com;
        }
    }
    return best_com;
}



int LouvainLevel::n2c(int node)
{
    return m_n2c[node];
}


void LouvainLevel::init(Graph * g)
{
    reset();
    m_g = g;
    int maxtime = m_g->maxtime();
    degree.resize(maxtime);
    init_alone();

    for(std::list< int >::const_iterator itnodes = m_g->nodes().begin(); itnodes != m_g->nodes().end(); itnodes++)
    {
        for(int time = 0; time < maxtime; time++)
        {
            degree[time][n2c(*itnodes)] = 0;
        }
    }

    for(int time = g->mintime(); time < maxtime; time++)
    {
        for(std::list< int >::const_iterator it = m_g->nodes().begin(); it != m_g->nodes().end(); it++)
        {
            int src = *it;
            if(m_g->has_node(src, time))
            {
                degree[time][n2c(src)] += m_g->degree(src, time);
            }
        }
    }
}



void LouvainLevel::reset()
{
    degree.clear();
    m_g = NULL;
    m_n2c.clear();
}

void LouvainLevel::init_alone()
{
    int count = 0;
    for(std::list< int >::const_iterator it = m_g->nodes().begin(); it != m_g->nodes().end(); it++)
    {
        m_n2c[*it] = count;
        count++;
    }
}